@@ -1,8 +1,8 @@
 ##---------------------------------------------------------------------------##
 ##  File:
-##	$Id: readmail.pl,v 2.33 2003/08/02 06:04:47 ehood Exp $
+##	$Id: readmail.pl,v 2.42 2011/01/09 05:13:14 ehood Exp $
 ##  Author:
-##      Earl Hood       mhonarc@mhonarc.org
+##      Earl Hood       mhonarc AT mhonarc DOT org
 ##  Description:
 ##      Library defining routines to parse MIME e-mail messages.  The
 ##	library is designed so it may be reused for other e-mail
@@ -24,7 +24,7 @@
 ##	$parm_hash_ref  = MAILparse_parameter_str($header_field, 1);
 ##
 ##---------------------------------------------------------------------------##
-##    Copyright (C) 1996-2002	Earl Hood, mhonarc@mhonarc.org
+##    Copyright (C) 1996-2003	Earl Hood, mhonarc AT mhonarc DOT org
 ##
 ##    This program is free software; you can redistribute it and/or modify
 ##    it under the terms of the GNU General Public License as published by
@@ -44,12 +44,13 @@
 
 package readmail;
 
+$DEBUG = 0;
+
 ###############################################################################
 ##	Private Globals							     ##
 ###############################################################################
 
-my $Url	          = '(\w+://|\w+:)';
-
+#my $Url	  = '(\w+://|\w+:)';
 my @_MIMEAltPrefs = ();
 my %_MIMEAltPrefs = ();
 
@@ -61,6 +62,18 @@ my %_MIMEAltPrefs = ();
 ##	Constants
 ##
 
+## String for matching the start of a URL: It seems unnecessary to
+#  try to recognize all valid schemes, so we use a simplier regex.
+#  Keep the old one around just in case we need to resurrect it.
+#$UrlRxStr	  =
+#    '(?:(?:https?|ftp|afs|wais|telnet|ldap|gopher|z39\.50[rs]|vemmi|imap|'.
+#          'nfs|acap|rtspu?|tip|pop|sip|(?:soap|xmlrpc)\.beeps?|go|ipp|'.
+#          'tftp)://|'.
+#       'news:(?://)?|'.
+#       '(?:nntp|mid|cid|mailto|prospero|data|service|tel|fax|modem|h\.323):)';
+$UrlRxStr	  =
+    '(?:(?:https?|ftp|ldap|gopher)://|news:(?://)?|(?:nntp|mailto):)';
+
 ##  Constants for use as second argument to MAILdecode_1522_str().
 sub JUST_DECODE() { 1; }
 sub DECODE_ALL()  { 2; }
@@ -104,9 +117,9 @@ $DecodeHeader	= 0;
 ##  set to true.
 
 %MIMEDecoders			= ()
-    unless defined(%MIMEDecoders);
+    unless %MIMEDecoders;
 %MIMEDecodersSrc		= ()
-    unless defined(%MIMEDecodersSrc);
+    unless %MIMEDecodersSrc;
 
 ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ##  %MIMECharSetConverters is the associative array for storing functions
@@ -140,9 +153,9 @@ $DecodeHeader	= 0;
 ##  string.
 
 %MIMECharSetConverters			= ()
-    unless defined(%MIMECharSetConverters);
+    unless %MIMECharSetConverters;
 %MIMECharSetConvertersSrc		= ()
-    unless defined(%MIMECharSetConvertersSrc);
+    unless %MIMECharSetConvertersSrc;
 
 ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ##  %MIMEFilters is the associative array for storing functions that
@@ -167,9 +180,9 @@ $DecodeHeader	= 0;
 ##  that all functions are defined before invoking MAILread_body.
 
 %MIMEFilters	= ()
-    unless defined(%MIMEFilters);
+    unless %MIMEFilters;
 %MIMEFiltersSrc	= ()
-    unless defined(%MIMEFiltersSrc);
+    unless %MIMEFiltersSrc;
 
 ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ##  %MIMEFiltersArgs is the associative array for storing any optional
@@ -183,7 +196,7 @@ $DecodeHeader	= 0;
 ##  listed for a function if both are applicable.
 
 %MIMEFiltersArgs	= ()
-    unless defined(%MIMEFiltersArgs);
+    unless %MIMEFiltersArgs;
 
 ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ##  %MIMEExcs is the associative array listing which data types
@@ -191,12 +204,23 @@ $DecodeHeader	= 0;
 ##
 ##	Keys => content-type, or base-type
 ##	Values => <should evaluate to a true expression>
-##
-##  For purposes of efficiency, content-types, or base-types, should
-##  be specified in lowercase.  All key lookups are done in lowercase.
 
 %MIMEExcs			= ()
-    unless defined(%MIMEExcs);
+    unless %MIMEExcs;
+
+## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+##  %MIMEIncs is the associative array listing which data types
+##  should be auto-included during parsing:
+##
+##	Keys => content-type, or base-type
+##	Values => <should evaluate to a true expression>
+##
+##  If there are any keys defined in %MIMEIncs, then any content-type
+##  not in the hash is automatically excluded.  I.e.  %MIMEIncs can
+##  be used to only allow a well-defined set of content-types.
+
+%MIMEIncs			= ()
+    unless %MIMEIncs;
 
 ## - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ##  %MIMECharsetAliases is a mapping of charset names to charset names.
@@ -207,7 +231,7 @@ $DecodeHeader	= 0;
 ##	Values => real charset name
 ##
 %MIMECharsetAliases = ()
-    unless defined(%MIMECharsetAliases);
+    unless %MIMECharsetAliases;
 
 ##---------------------------------------------------------------------------
 ##	Text entity-related variables
@@ -222,10 +246,10 @@ $TextEncode = undef
     unless defined($TextEncode);
 ##  Text encoding function.
 $TextEncoderFunc = undef
-    unless defined($TextEncodingFunc);
+    unless defined($TextEncoderFunc);
 ##  Text encoding function source file.
 $TextEncoderSrc = undef
-    unless defined($TextEncodingSrc);
+    unless defined($TextEncoderSrc);
 
 ##  Prefilter function
 $TextPreFilter  = undef
@@ -377,7 +401,8 @@ sub MAILdecode_1522_str {
 			    ? $MIMECharsetAliases{$charset} : $charset;
 	    $strtxt =~ s/_/ /g;
 	    $strtxt =  &$dec($strtxt);
-	    &$encfunc(\$strtxt, $charset, $TextEncode);
+            $strtxt =~ s/[\r\n]/ /g;
+	    &$encfunc(\$strtxt, $real_charset, $TextEncode);
 	    $ret   .= $strtxt;
 
 	# Regular conversion
@@ -391,7 +416,9 @@ sub MAILdecode_1522_str {
 	    # Decode only
 	    if ($charcnv eq '-decode-') {
 		$strtxt =~ s/_/ /g;
-		$ret .= &$dec($strtxt);
+		$strtxt =  &$dec($strtxt);
+                $strtxt =~ s/[\r\n]/ /g;
+		$ret .= $strtxt;
 
 	    # Ignore if just decoding
 	    } elsif ($dec_flag) {
@@ -400,7 +427,9 @@ sub MAILdecode_1522_str {
 	    # Decode and convert
 	    } elsif (defined(&$charcnv)) {
 		$strtxt =~ s/_/ /g;
-		$ret .= &$charcnv(&$dec($strtxt), $real_charset);
+		$strtxt =  &$dec($strtxt);
+                $strtxt =~ s/[\r\n]/ /g;
+		$ret .= &$charcnv($strtxt, $real_charset);
 
 	    # Fallback is to ignore
 	    } else {
@@ -428,7 +457,7 @@ sub MAILdecode_1522_str {
 ##	MAILread_body() parses a MIME message body.
 ##	Usage:
 ##	  ($data, @files) =
-##	      MAILread_body($fields_hash_ref, $body_date_ref);
+##	      MAILread_body($fields_hash_ref, $body_data_ref);
 ##
 ##	Parameters:
 ##	  $fields_hash_ref
@@ -541,10 +570,17 @@ sub MAILread_body {
 	$fields->{'x-mha-charset'} = $charset;
 	my $textfunc = load_textencoder();
 	if (defined($textfunc)) {
+            if ($DEBUG) {
+              print STDERR "MAILread_body: have textfunc: $textfunc\n";
+            }
 	    $fields->{'x-mha-charset'} = $TextEncode
 		if defined(&$textfunc($body, $charset, $TextEncode));
 	}
 	if (defined($TextPreFilter) && defined(&$TextPreFilter)) {
+            if ($DEBUG) {
+              print STDERR 'MAILread_body: have TextPreFilter: ',
+                           $TextPreFilter, "\n";
+            }
 	    &$TextPreFilter($fields, $body);
 	}
     } else {
@@ -552,6 +588,9 @@ sub MAILread_body {
 	# a non-text type
 	$fields->{'x-mha-charset'} = $TextDefCharset;
     }
+    if ($DEBUG) {
+      print STDERR 'MAILread_body: charset: ',$fields->{'x-mha-charset'},"\n";
+    }
 
     ## A filter is defined for given content-type
     if ($filter && defined(&$filter)) {
@@ -820,8 +859,6 @@ sub MAILread_header {
     my $header = '';
     my($value, $tmp, $pos);
 
-    my $encfunc = load_textencoder();
-
     ## Read a line at a time.
     for ($pos=0; $pos >= 0; ) {
 	$pos = index($$mesg, "\n");
@@ -838,13 +875,6 @@ sub MAILread_header {
 	    $header .= $tmp;
 	}
 
-	## Decode text if requested
-	if (defined($encfunc)) {
-	    $tmp = &MAILdecode_1522_str($tmp,TEXT_ENCODE);
-	} elsif ($DecodeHeader) {
-	    $tmp = &MAILdecode_1522_str($tmp,JUST_DECODE);
-	}
-
 	## Check for continuation of a field
 	if ($tmp =~ /^\s/) {
 	    $fields->{$label}[-1] .= $tmp  if $label;
@@ -861,6 +891,7 @@ sub MAILread_header {
 	    }
 	}
     }
+    decode_1522_fields($fields);
     ($fields, $header);
 }
 
@@ -875,8 +906,6 @@ sub MAILread_file_header {
     my $handle = shift;
     my $encode = shift;
 
-    my $encfunc = load_textencoder();
-
     my $label  = '';
     my $header = '';
     my $fields = { };
@@ -890,13 +919,6 @@ sub MAILread_file_header {
 	## Delete eol characters
 	$tmp =~ s/[\r\n]//g;
 
-	## Decode text if requested
-	if (defined($encfunc)) {
-	    $tmp = &MAILdecode_1522_str($tmp,TEXT_ENCODE);
-	} elsif ($DecodeHeader) {
-	    $tmp = &MAILdecode_1522_str($tmp,JUST_DECODE);
-	}
-
 	## Check for continuation of a field
 	if ($tmp =~ /^\s/) {
 	    $fields->{$label}[-1] .= $tmp  if $label;
@@ -913,6 +935,7 @@ sub MAILread_file_header {
 	    }
 	}
     }
+    decode_1522_fields($fields);
     ($fields, $header);
 }
 
@@ -922,14 +945,28 @@ sub MAILread_file_header {
 ##
 sub MAILis_excluded {
     my $ctype = lc($_[0]) || 'text/plain';
-    if ($MIMEExcs{$ctype}) {
-	return 1;
-    }
-    if ($ctype =~ s/\/x-/\//) {
-	return 1  if $MIMEExcs{$ctype};
-    }
+    my $btype = undef;
+
+    $ctype =~ s/\/x-/\//;
     if ($ctype =~ m|([^/]+)/|) {
-	return $MIMEExcs{$1};
+	$btype = $1;
+    }
+
+    MIMEINCS: {
+	# Treat multipart special: It is always included unless present
+	# in MIMEExcs.
+	last  MIMEINCS  if ($ctype =~ /^multipart\b/);
+
+	if (%MIMEIncs) {
+	    if ($MIMEIncs{$ctype} || (defined($btype) && $MIMEIncs{$btype})) {
+		last MIMEINCS;
+	    } else {
+		return 1;
+	    }
+	}
+    }
+    if ($MIMEExcs{$ctype} || (defined($btype) && $MIMEExcs{$btype})) {
+	return 1;
     }
     0;
 }
@@ -1250,37 +1287,76 @@ sub extract_ctype {
 }
 
 ##---------------------------------------------------------------------------##
-
+##	apply_base_url(): Convert a relative URL to a full URL with
+##	specific base;
+##
 sub apply_base_url {
-    my($b, $u) = @_;
+    my $b = shift;  # Base URL
+    my $u = shift;  # URL to apply base to
+
+    ## If no base, nothing to do
     return $u  if !defined($b) || $b !~ /\S/;
 
-    my($ret);
+    ## If absolute URL or scroll link; do nothing
     $u =~ s/^\s+//;
-    if ($u =~ m%^$Url%o || $u =~ m/^#/) {
-	## Absolute URL or scroll link; do nothing
-        $ret = $u;
-    } else {
-	## Relative URL
-	if ($u =~ /^\./) {
-	    ## "./---" or "../---": Need to remove and adjust base
-	    ## accordingly.
-	    $b =~ s/\/$//;
-	    my @a = split(/\//, $b);
-	    my $cnt = 0;
-	    while ( $cnt <= scalar(@a) &&
-		    $u =~ s|^(\.{1,2})/|| ) { ++$cnt  if length($1) == 2; }
-	    splice(@a, -$cnt)  if $cnt > 0;
-	    $b = join('/', @a, "");
-
-	} elsif ($u =~ m%^/%) {
-	    ## "/---": Just use hostname:port of base.
-	    $b =~ s%^(${Url}[^/]*)/.*%$1%o;
-	}
-        $ret = $b . $u;
+    if ($u =~ /^$UrlRxStr/o || $u =~ m/^#/) {
+        return $u;
     }
-    $ret;
+
+    ## Check if base URL allows relative resolution
+    my($host_part, $scheme);
+    if ($b =~ s{^((https?|ftp|file|nfs|acap|tftp)://[\w\-:\d.\@\%=~&]+)/?}{}) {
+	$host_part = $1;
+	$scheme    = lc $2;
+    } else {
+	warn qq/Warning: Invalid base url, "$b" to apply to "$u"\n/;
+	return $u;
+    }
+
+    ## If "/---", just use hostname:port of base.
+    if ($u =~ /^\//) {
+	return $host_part . $u;
+    }
+
+    ## Clean up base URL
+    SCHEME: {
+	if ($scheme eq 'http' || $scheme eq 'https' || $scheme eq 'acap') {
+	    $b =~ s/\?.*$//;
+	    last SCHEME;
+	}
+	if ($scheme eq 'ftp') {
+	    $b =~ s/;type=.$//;
+	    last SCHEME;
+	}
+	if ($scheme eq 'tftp') {
+	    $b =~ s/;mode=\w+$//;
+	    last SCHEME;
+	}
+    }
+    $b =~ s/\/$//;    # strip any trailing '/' (we add it back later)
+
+    ## "./---" or "../---": Need to remove and adjust base accordingly.
+    my $cnt = 0;
+    while ( $u =~ s|^(\.{1,2})/|| ) { ++$cnt  if length($1) == 2; }
+
+    if ($b eq '') {
+	# base is just host
+	return join('/', $host_part, $u);
+    }
+    if ($cnt > 0) {
+	# trim path
+	my @a = split(/\//, $b);
+	if ($cnt <= scalar(@a)) {
+	    splice(@a, -$cnt);
+	    return join('/', $host_part, @a, $u);
+	}
+	# invalid relative path, tries to go past root
+	return join('/', $host_part, $u);
+
+    }
+    return join('/', $host_part, $b, $u);
 }
+
 ##---------------------------------------------------------------------------##
 
 sub extract_charset {
@@ -1297,7 +1373,7 @@ sub extract_charset {
     # If HTML, check <meta http-equiv=content-type> tag since it
     # can be different than what is specified in the entity header.
     if (($subtype eq 'html' || $subtype eq 'x-html') &&
-	($body =~ m/(<meta\s+http-equiv\s*=\s*['"]?
+	($$body =~ m/(<meta\s+http-equiv\s*=\s*['"]?
 		     content-type\b[^>]*>)/xi)) {
 	my $meta = $1;
 	if ($meta =~ m/\bcharset\s*=\s*['"]?([\w\.\-]+)/i) {
@@ -1333,6 +1409,25 @@ sub gen_full_part_number {
 
 ##---------------------------------------------------------------------------##
 
+sub decode_1522_fields {
+  my $fields  = shift;
+  my $encfunc = load_textencoder();
+  my $mode    = defined($encfunc)
+                  ? TEXT_ENCODE : ($DecodeHeader ? JUST_DECODE : -1);
+  if ($mode == -1) {
+    return $fields;
+  }
+  my ($label, $value, $v);
+  while (($label, $value) = each %$fields) {
+    foreach $v (@$value) {
+      $v = &MAILdecode_1522_str($v, $mode);
+    }
+  }
+  $fields;
+}
+
+##---------------------------------------------------------------------------##
+
 sub dump_header {
     my $fh	= shift;
     my $fields	= shift;
